/************************************************************
 **
 ** picmovlpop.c : PIC data optimization definitions
 **
 ** Copyright (c) 2010, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/
#include <assert.h>
#include "../libcore/pf_proc.h"
#include "pic_msg.h"
#include "piccolst.h"
#include "pic_gop.h"
#include "pic_brop.h"
#include "picmovlpop.h"

typedef struct pic_code_movlp_data_ {
  uchar pclath[PIC_CODE_GOP_DEPTH];
  unsigned ix;
} pic_code_movlp_data_t;

/*
 * NAME
 *   pic_code_movlp_analyze
 *
 * DESCRIPTION
 *   analyze the movlp in the pic_codes[] array
 *
 * PARAMETERS
 *   pf    : 
 *   pcl   : pcl on entry
 *   pclath: current contents of pclath
 *   depth : depth (for statistics only)
 *
 * RETURN
 *   none
 *
 * NOTES
 */
static pic_code_gop_rc_t pic_code_movlp_analyze(pfile_t *pf, 
  pic_code_t code, pic_code_gop_which_t which, void *odata)
{
  pic_code_gop_rc_t      rc;
  pic_code_movlp_data_t *data;
  uchar                  pclath;

  data = odata;
  if (PIC_CODE_NONE != code) {
    pclath = data->pclath[data->ix];
  }
  rc = PIC_CODE_GOP_RC_CONTINUE;
  switch (which) {
    case PIC_CODE_GOP_WHICH_ENTER:
      data->ix++;
      assert(data->ix < PIC_CODE_GOP_DEPTH);
      if (data->ix > 0) {
        data->pclath[data->ix] = data->pclath[data->ix - 1];
      }
      break;
    case PIC_CODE_GOP_WHICH_EXIT:
      data->ix--;
      break;
    case PIC_CODE_GOP_WHICH_PRE:
      if ((pic_code_pclath_get(code) == pclath) 
        || (PIC_PCLATH_INDETERMINITE == pic_code_pclath_get(code))) {
        rc = PIC_CODE_GOP_RC_STOP;
      } else {
        if (PIC_PCLATH_UNKNOWN == pic_code_pclath_get(code)) {
          pic_code_pclath_set(code, pclath);
        } else if (pclath != pic_code_pclath_get(code)) {
          pic_code_pclath_set(code, PIC_PCLATH_INDETERMINITE);
        }
        switch (pic_code_op_get(code)) {
          case PIC_OPCODE_MOVLP:
            {
              label_t lbl;

              lbl = pic_code_brdst_get(code);
              pclath = label_pc_get(lbl) >> 8;
            }
            break;
          default:
            break;
        }
      }
      break;
    case PIC_CODE_GOP_WHICH_POST:
      if (pic_code_is_suspend(pf, code)
        || (PIC_OPCODE_CALL == pic_code_op_get(code))) {
        pclath = PIC_PCLATH_INDETERMINITE;
      }
      break;
  }
  if (PIC_CODE_NONE != code) {
    data->pclath[data->ix] = pclath;
  }
  return rc;
}

static void pic_code_movlp_opt(pfile_t *pf,
    pic_code_gop_who_t who, void *odata)
{
  pic_code_movlp_data_t *data;

  UNUSED(pf);
  data = odata;
  switch (who) {
    case PIC_CODE_GOP_WHO_ISR:
    case PIC_CODE_GOP_WHO_INDIRECT:
      data->pclath[0] = PIC_PCLATH_INDETERMINITE;
      break;
    case PIC_CODE_GOP_WHO_USER:
      data->pclath[0] = 0;
      break;
  }
  data->ix = (unsigned) -1;
}

/* run through the pic_codes and remove any redundant dataset commands */
static unsigned pic_code_pclath_post_analyze(pfile_t *pf)
{
  pic_code_t pv;
  pic_code_t code;
  unsigned   ins_ct; /* instruction ct */
  unsigned   rem_ct;

  for (ins_ct = 0, rem_ct = 0, pv = 0, code = pic_code_list_head_get(pf);
       code;
       ins_ct++) {
    pic_code_t next;

    next = pic_code_next_get(code);
    if (!pic_code_flag_test(code, PIC_CODE_FLAG_NO_OPTIMIZE)) {
      boolean_t            cremove;

      switch (pic_code_op_get(code)) {
        case PIC_OPCODE_MOVLP:
          cremove = (label_pc_get(pic_code_brdst_get(code)) >> 8)
              == pic_code_pclath_get(code);
          break;
        default:
          cremove = BOOLEAN_FALSE;
          break;
      }
      if (cremove) {
        pic_code_list_remove(pf, code);
        pic_code_free(code);
        rem_ct++;
      } else {
        pv = code;
      }
    }
    code = next;
  }
  pfile_log(pf, PFILE_LOG_DEBUG, "...removed %u of %u instructions", 
      rem_ct, ins_ct);
  return rem_ct;
}

void pic_code_movlp_optimize(pfile_t *pf)
{
  pic_code_movlp_data_t data;

  pfile_log(pf, PFILE_LOG_DEBUG, PIC_MSG_FIXING_DATA_BITS);
  do {
    pic_branchbits_pc_set(pf);
    pic_code_gop_optimize(pf, pic_code_movlp_opt,
      pic_code_movlp_analyze, &data);
  } while (pic_code_pclath_post_analyze(pf));
}

